import argparse
import random
import string
import torch.nn.functional
import numpy as np
import torch
import torch.utils.data
import torchvision

# pip3 install multicolorcaptcha
from multicolorcaptcha import CaptchaGenerator

from matplotlib import pyplot as plt
from torch.nn.utils.rnn import pack_padded_sequence, PackedSequence
from tqdm import tqdm

from english_words import english_words_lower_alpha_set
english_words_lower_alpha_set = list(english_words_lower_alpha_set)

# up to 4 characters long words
english_words_lower_alpha_set = [word for word in english_words_lower_alpha_set if 2 <= len(word) <= 4]

plt.rcParams["figure.figsize"] = (15, 5)
plt.style.use('dark_background')

parser = argparse.ArgumentParser()

parser.add_argument('-device', type=str, default='cpu')
parser.add_argument('-batch_size', type=int, default=24)
parser.add_argument('-learning_rate', type=float, default=1e-3)


parser.add_argument('-num_epochs', type=int, default=100)
parser.add_argument('-input_captcha_samples_per_epoch', type=int, default=1000)

args = parser.parse_args()

class Dataset(torch.utils.data.Dataset):
    def __init__(self):
        super().__init__()
        self.characters = ' ' # blank ' '
        for i in range(0, 10):
            self.characters += str(i)
        self.characters += '#' # End of word
        self.eos_idx = self.characters.find('#')

        self.input_captcha_length = max([len(it) for it in english_words_lower_alpha_set]) + 1

    def __len__(self):
        return args.input_captcha_samples_per_epoch

    def __getitem__(self, idx):

        generator = CaptchaGenerator(captcha_size_num=1)
        captcha = generator.gen_captcha_image(
            difficult_level=1,
            chars_mode="nums",
            margin=False
        )

        y_str = captcha.characters
        y_str += '#' # after generator add eos token
        y_idxes = np.array([self.characters.find(x) for x in y_str])

        x = np.array(captcha.image) / 255.0
        x = x.swapaxes(0, 2).astype(np.float32)

        y_idxes_padded = np.zeros((self.input_captcha_length,), dtype=np.long)
        y_idxes_padded[:len(y_str)] = y_idxes

        return x, y_idxes_padded, len(y_str)


dataset_full = Dataset()

dataloader_train = torch.utils.data.DataLoader(
    dataset=dataset_full,
    batch_size=args.batch_size,
    shuffle=True,
    drop_last=(len(dataset_full) % args.batch_size < args.batch_size * 0.7)
)

# for i in range(0, 10):
#     x_img = dataset_full[i][0].swapaxes(0, 2)
#     print(x_img.shape)
#     plt.imshow(x_img)
#     plt.show()
# exit()

class Model(torch.nn.Module):
    def __init__(self):
        super().__init__()

        # TODO

    def forward(self, x):
        # TODO

        return y_prim


model = Model()
model = model.to(args.device)
optimizer = torch.optim.Adam(model.parameters(), lr=args.learning_rate)

losses_all = []
accs_all = []

for epoch in range(1, args.num_epochs + 1):
    losses = []
    accs = []
    for x, y_idxes, y_len in tqdm(dataloader_train, desc=f'epoch: {epoch}'):
        x = x.to(args.device)

        y_prim = model.forward(x)

        y_prim_len = []
        for each in y_prim:
            each_len = len(each)
            if dataset_full.eos_idx in each:
                each_len = each.index(dataset_full.eos_idx) + 1
            y_prim_len.append(each_len)
        y_prim_len = torch.LongTensor(np.array(y_prim_len))

        y_prim_packed = pack_padded_sequence(y_prim, y_len, batch_first=True, enforce_sorted=False)
        y_idxes_packed = pack_padded_sequence(y_idxes, y_len, batch_first=True, enforce_sorted=False)

        loss = torch.nn.functional.cross_entropy(y_prim_packed.data, y_idxes_packed.data)

        loss.backward()
        optimizer.step()
        optimizer.zero_grad()

        y_prim_idxes_packed = y_prim_packed.data.argmax(-1)
        y_prim_idxes = y_prim.argmax(-1)

        acc = torch.mean(torch.sum(torch.eq(y_prim_idxes_packed, y_idxes_packed.data).float(), dim=-1))

        accs.append(acc.item())
        losses.append(loss.item())

        # save model at best loss or accuracy for validation
        if losses[-1] == min(losses):
            torch.save(model.cpu().state_dict(), f'model_epoch_{epoch}.pt')

    loss_epoch = np.mean(losses)
    losses_all.append(loss_epoch)

    acc_epoch = np.mean(accs)
    accs_all.append(acc_epoch)

    print(f'epoch {epoch}: loss {loss_epoch} acc: {acc_epoch}')

    plt.clf()
    plt.subplot(121)  # row col idx
    plts = []
    c = 0
    for key, value in [
        ('loss', losses_all),
        ('acc', accs_all)
    ]:
        plts += plt.plot(value, f'C{c}', label=key)
        ax = plt.twinx()
        c += 1

    plt.legend(plts, [it.get_label() for it in plts])

    for i, j in enumerate([4, 5, 6, 10, 11, 12, 16, 17, 18]):
        plt.subplot(3, 6, j)

        plt.imshow(x[i].cpu().numpy().swapaxes(0, 2))
        y_str = ''.join([dataset_full.characters[it] for it in y_idxes[i].cpu().numpy()])
        y_str = y_str.replace(' ', '')
        if '#' in y_str:
            y_str = y_str[:y_str.index('#')]

        y_prim_str = ''.join([dataset_full.characters[it] for it in y_prim_idxes[i].cpu().numpy()])
        y_prim_str = y_prim_str.replace(' ', '')
        if '#' in y_prim_str:
            y_prim_str = y_prim_str[:y_prim_str.index('#')]
        plt.title(y_str + '\n' + y_prim_str)

    plt.tight_layout(pad=0.5)
    plt.show()
